"""Infer types for AST nodes.
"""
from ._ass import Ass
from ._ast import find_node
from ._handlers import get_type
from ._type import Type


__version__ = '0.2.0'
__all__ = ['find_node', 'get_type', 'Ass', 'Type']



import ast
from collections import deque
from inspect import getdoc
from typing import Iterator

import astroid


def walk(module: astroid.Module) -> Iterator[astroid.NodeNG]:
    stack = deque([module])
    while stack:
        node = stack.pop()
        doc_node = getattr(node, 'doc_node', None)
        if doc_node is not None and doc_node is not node:
            stack.append(doc_node)
        stack.extend(node.get_children())
        yield node


def find_node(module: astroid.Module, ast_node: ast.AST) -> astroid.NodeNG:
    if isinstance(ast_node, ast.Module):
        return module
    matches = []
    for node in walk(module):
        if node.lineno != ast_node.lineno:
            continue
        if node.col_offset != ast_node.col_offset:
            continue
        matches.append(node)
    if len(matches) == 1:
        return matches[0]

    if not matches:
        if isinstance(ast_node, ast.Expr):
            return find_node(module, ast_node.value)
        # different version of Python point either at the function start
        # or at the first decorator. Astroid is consistent.
        if isinstance(ast_node, ast.FunctionDef):
            for node in walk(module):
                if not isinstance(node, astroid.FunctionDef):
                    continue
                if node.name == ast_node.name:
                    return node

    node_name = type(ast_node).__name__
    doc_entry = f':class:`ast.{node_name}`'
    for match in matches:
        doc = getdoc(type(match))
        if doc is not None and doc_entry in doc:
            return match
    if matches:
        return matches[0]
    msg = f'no matches for {type(ast_node).__name__} node '
    msg += f'at {ast_node.lineno}:{ast_node.col_offset}'
    raise LookupError(msg)



from __future__ import annotations
import pytest
from astypes import Type

t = Type.new


def u(*args):
    new_args = []
    for arg in args:
        if isinstance(arg, str):
            arg = t(arg)
        new_args.append(arg)
    return t('Union', args=new_args)


@pytest.mark.parametrize('left, right, result', [
    # simplify the same type twice
    ('float',   'float',    'float'),
    ('str',     'str',      'str'),

    # simplify int+float
    ('float',   'int',      'float'),
    ('int',     'float',    'float'),
    ('int',     'int',      'int'),

    # simplify one unknown
    ('',    'int',  'int'),
    ('int', '',     'int'),
    ('',    '',     ''),

    # actual unions
    ('str',  'int',  u('str', 'int')),
    ('int',  'str',  u('int', 'str')),
    ('int',  'None', u('int', 'None')),
    ('None', 'int',  u('int', 'None')),
])
def test_merge(left, right, result):
    if isinstance(left, str):
        left = t(left)
    if isinstance(right, str):
        right = t(right)
    if isinstance(result, str):
        result = t(result)
    assert left.merge(right) == result


@pytest.mark.parametrize('given, expected', [
    (t('int'),                      'int'),
    (u('int', 'str'),               'int | str'),
    (t('list', args=[t('int')]),    'list[int]'),
    (t(''),                         'Any'),
])
def test_annotation(given: Type, expected: str):
    assert given.annotation == expected



import astroid
import pytest
from astypes import get_type


@pytest.mark.parametrize('expr, type', [
    # literals
    ('1',       'int'),
    ('1.2',     'float'),
    ('"hi"',    'str'),
    ('f"hi"',   'str'),
    ('b"hi"',   'bytes'),
    ('""',      'str'),
    ('None',    'None'),
    ('True',    'bool'),

    # collection literals
    ('[]',      'list'),
    ('[1]',     'list'),
    ('()',      'tuple'),
    ('(1,)',    'tuple'),
    ('{}',      'dict'),
    ('{1:2}',   'dict'),
    ('{1,2}',   'set'),

    # collection constructors
    ('list()',      'list'),
    ('list(x)',     'list'),
    ('dict()',      'dict'),
    ('dict(x)',     'dict'),
    ('set()',       'set'),
    ('set(x)',      'set'),
    ('tuple()',     'tuple'),
    ('tuple(x)',    'tuple'),

    # other type constructors
    ('int()',       'int'),
    ('int(x)',      'int'),
    ('str()',       'str'),
    ('str(x)',      'str'),
    ('float()',     'float'),
    ('float(x)',    'float'),

    # math operations
    ('3 + 2',       'int'),
    ('3 * 2',       'int'),
    ('3 + 2.',      'float'),
    ('3. + 2',      'float'),
    ('3 / 2',       'float'),
    ('"a" + "b"',   'str'),

    # operations with known type
    ('not x',       'bool'),
    ('x is str',    'bool'),
    ('x and y',     'bool'),
    ('x or y',      'bool'),

    # operations with assumptions
    ('x in (1, 2, 3)',  'bool'),
    ('x < 10',          'bool'),
    ('~13',             'int'),
    ('+13',             'int'),

    # methos of builtins
    ('"".join(x)',      'str'),
    ('[1,2].count(1)',  'int'),

    # builtin functions
    ('len(x)',          'int'),

    # comprehensions
    ('[x for x in y]',      'list'),
    ('{x for x in y}',      'set'),
    ('{x: y for x in z}',   'dict'),
    ('(x for x in y)',      'Iterator'),

    # misc
    ('Some(x)',             'Some'),
])
def test_expr(expr, type):
    node = astroid.extract_node(f'None\n{expr}')
    t = get_type(node)
    assert t is not None
    assert t.annotation == type


@pytest.mark.parametrize('expr', [
    'min(x)',
    'x',
    'X',
    'WAT',
    'wat()',
    'WAT()',
    '+x',
    'x + y',
    '1 + y',
    'x + 1',
    '"a" + 1',
    'str.wat',
    '"hi".wat',
    'None.hi',
    'None.hi()',
    '"hi".wat()',
    'wat.wat',
    'super().something()',
])
def test_cannot_infer_expr(expr):
    node = astroid.extract_node(expr)
    assert get_type(node) is None


@pytest.mark.parametrize('setup, expr, type', [
    ('import math',             'math.sin(x)',  'float'),
    ('from math import sin',    'sin(x)',       'float'),
    ('my_list = list',          'my_list(x)',   'list'),
])
def test_astroid_inference(tmp_path, setup, expr, type):
    node = astroid.extract_node(f'{setup}\n{expr}')
    t = get_type(node)
    assert t is not None
    assert t.annotation == type



from __future__ import annotations

from enum import Enum


class Ass(Enum):
    """Assumptions about the types that might be not true but usually are true.
    """
    # cannot infer type of one or more of the return statements,
    # assume all return statements to have the same type.
    # It is here for the sake of infer-types.
    ALL_RETURNS_SAME = 'all-returns-same'
    # assume that comparison operations aren't overloaded
    NO_COMP_OVERLOAD = 'no-comp-overload'
    # assume that unary operators aren't overloaded
    NO_UNARY_OVERLOAD = 'no-unary-overload'
    # assume that all CamelCase names are types
    CAMEL_CASE_IS_TYPE = 'camel-case-is-type'
    # assume that built-in types and functions aren't shadowed
    NO_SHADOWING = 'camel-case-is-type'



from __future__ import annotations

import astroid

from ._type import Type


def infer(node: astroid.NodeNG) -> list:
    try:
        return list(node.infer())
    except astroid.InferenceError:
        return []


def qname_to_type(qname: str) -> Type:
    if qname.startswith('builtins.'):
        qname = qname.split('.')[-1]
    if qname == 'NoneType':
        qname = 'None'
    if '.' not in qname:
        return Type.new(qname)
    mod_name, _, obj_name = qname.rpartition('.')
    return Type.new(obj_name, module=mod_name)


def is_camel(name: str) -> bool:
    if not name:
        return False
    if not name[0].isupper():
        return False
    if not any(c.islower() for c in name):
        return False
    return True



import ast
from pathlib import Path
import astroid
import pytest
import astypes

ROOT = Path(astypes.__path__[0])
paths = [pytest.param(p, id=p.name) for p in ROOT.glob('*.py')]
UNSUPPORTED = (
    ast.alias, ast.expr_context,
    ast.arguments, ast.cmpop, ast.operator,
    ast.boolop, ast.unaryop, ast.comprehension,
)


@pytest.mark.parametrize('path', sorted(paths))
def test_smoke(path: Path):
    source = path.read_text()
    astroid_tree = astroid.parse(source)
    ast_tree = ast.parse(source)
    for ast_node in ast.walk(ast_tree):
        if isinstance(ast_node, UNSUPPORTED):
            continue
        astroid_node = astypes.find_node(astroid_tree, ast_node)
        if hasattr(ast_node, 'lineno'):
            assert astroid_node.lineno <= ast_node.lineno



from __future__ import annotations

import ast
from dataclasses import dataclass, field
from logging import getLogger
from typing import Callable, TypeVar

import astroid
import typeshed_client

from ._ass import Ass
from ._helpers import infer, is_camel, qname_to_type
from ._type import Type


logger = getLogger(__package__)
Handler = Callable[[astroid.NodeNG], 'Type | None']
T = TypeVar('T', bound=Handler)


@dataclass
class Handlers:
    _registry: list[tuple[type, Handler]] = field(default_factory=list)

    def get_type(self, node: astroid.NodeNG) -> Type | None:
        """Infer type of the given astroid node.

        If the type cannot be inferred, None is returned.
        Keep in mind that a number of assumptions can be made about the code
        in order to infer the type. Use `Type.assumptions` to see them.
        """
        for supported_type, handler in self._registry:
            if isinstance(node, supported_type):
                result = handler(node)
                if result is not None:
                    return result
        return None

    def register(self, t: type) -> Callable[[T], T]:
        def callback(handler):
            self._registry.append((t, handler))
            return handler
        return callback


handlers = Handlers()
get_type = handlers.get_type


@handlers.register(astroid.Const)
def _handle_const(node: astroid.Const) -> Type | None:
    if node.value is None:
        return Type.new('None')
    return Type.new(type(node.value).__name__)


@handlers.register(astroid.JoinedStr)
def _handle_fstring(node: astroid.JoinedStr) -> Type | None:
    return Type.new('str')


@handlers.register(astroid.List)
def _handle_list(node: astroid.List) -> Type | None:
    return Type.new('list')


@handlers.register(astroid.Tuple)
def _handle_tuple(node: astroid.Tuple) -> Type | None:
    return Type.new('tuple')


@handlers.register(astroid.Dict)
def _handle_dict(node: astroid.Dict) -> Type | None:
    return Type.new('dict')


@handlers.register(astroid.Set)
def _handle_set(node: astroid.Set) -> Type | None:
    return Type.new('set')


@handlers.register(astroid.UnaryOp)
def _handle_unary_op(node: astroid.UnaryOp) -> Type | None:
    if node.op == 'not':
        return Type.new('bool')
    result = get_type(node.operand)
    if result is not None:
        result = result.add_ass(Ass.NO_UNARY_OVERLOAD)
        return result
    return None


@handlers.register(astroid.BinOp)
def _handle_binary_op(node: astroid.BinOp) -> Type | None:
    assert node.op
    lt = get_type(node.left)
    if lt is None:
        return None
    rt = get_type(node.right)
    if rt is None:
        return None
    if lt.name == rt.name == 'int':
        if node.op == '/':
            return Type.new('float')
        return lt
    if lt.name in ('float', 'int') and rt.name in ('float', 'int'):
        return Type.new('float')
    if lt.name == rt.name:
        return rt
    return None


@handlers.register(astroid.BoolOp)
def _handle_bool_op(node: astroid.BoolOp) -> Type | None:
    return Type.new('bool')


@handlers.register(astroid.Compare)
def _handle_compare(node: astroid.Compare) -> Type | None:
    if node.ops[0][0] == 'is':
        return Type.new('bool')
    return Type.new('bool', ass={Ass.NO_COMP_OVERLOAD})


@handlers.register(astroid.ListComp)
def _handle_list_comp(node: astroid.ListComp) -> Type | None:
    return Type.new('list')


@handlers.register(astroid.SetComp)
def _handle_set_comp(node: astroid.SetComp) -> Type | None:
    return Type.new('set')


@handlers.register(astroid.DictComp)
def _handle_dict_comp(node: astroid.DictComp) -> Type | None:
    return Type.new('dict')


@handlers.register(astroid.GeneratorExp)
def _handle_gen_expr(node: astroid.GeneratorExp) -> Type | None:
    return Type.new('Iterator', module='typing')


@handlers.register(astroid.Call)
def _handle_call(node: astroid.Call) -> Type | None:
    if isinstance(node.func, astroid.Attribute):
        result = _get_attr_call_type(node.func)
        if result is not None:
            return result
    if isinstance(node.func, astroid.Name):
        _, symbol_defs = node.func.lookup(node.func.name)
        mod_name = 'builtins'
        if symbol_defs:
            symbol_def = symbol_defs[0]
            if isinstance(symbol_def, astroid.ImportFrom):
                mod_name = symbol_def.modname
        result = _get_ret_type_of_fun(mod_name, node.func.name)
        if result is not None:
            return result
        if is_camel(node.func.name):
            return Type.new(node.func.name, ass={Ass.CAMEL_CASE_IS_TYPE})
    return None


@handlers.register(astroid.NodeNG)
def _handle_infer_any(node: astroid.NodeNG) -> Type | None:
    for def_node in infer(node):
        if not isinstance(def_node, astroid.Instance):
            continue
        return qname_to_type(def_node.pytype())
    return None


@handlers.register(astroid.Call)
def _handle_call_infer(node: astroid.Call) -> Type | None:
    for def_node in infer(node.func):
        if not isinstance(def_node, astroid.FunctionDef):
            continue
        mod_name, _, fun_name = def_node.qname().rpartition('.')
        return _get_ret_type_of_fun(mod_name, fun_name)
    return None


def _get_ret_type_of_fun(
    mod_name: str,
    fun_name: str,
) -> Type | None:
    """For the given module and function name, get return type of the function.
    """
    module = typeshed_client.get_stub_names(mod_name)
    if module is None:
        logger.debug(f'no typeshed stubs for module {mod_name}')
        return None
    fun_def = module.get(fun_name)
    if fun_def is None:
        logger.debug('no typeshed stubs for module')
        return None
    if not isinstance(fun_def.ast, ast.FunctionDef):
        logger.debug('resolved call target is not a function')
        return None
    ret_node = fun_def.ast.returns
    return _conv_node_to_type(mod_name, ret_node)


def _get_attr_call_type(node: astroid.Attribute) -> Type | None:
    expr_type = get_type(node.expr)
    if expr_type is None:
        logger.debug('cannot get type of the left side of attribute')
        return None
    module = typeshed_client.get_stub_names('builtins')
    assert module is not None
    try:
        method_def = module[expr_type.name].child_nodes[node.attrname]
    except KeyError:
        logger.debug('not a built-in function')
        return None
    if not isinstance(method_def.ast, ast.FunctionDef):
        logger.debug('resolved call target of attr is not a function')
        return None
    ret_node = method_def.ast.returns
    return _conv_node_to_type('builtins', ret_node)


def _conv_node_to_type(
    mod_name: str,
    node: ast.AST | None,
) -> Type | None:
    import builtins
    import typing

    if node is None:
        logger.debug('no return type annotation for called function def')
        return None

    # fore generics, keep it generic
    if isinstance(node, ast.Subscript):
        return _conv_node_to_type(mod_name, node.value)

    # for regular name, check if it is a typing primitive or a built-in
    if isinstance(node, ast.Name):
        name = node.id
        if hasattr(builtins, name):
            return Type.new(name, ass={Ass.NO_SHADOWING})
        if name in typing.__all__:
            return Type.new(name, module='typing')
        logger.debug(f'cannot resolve {name} into a known type')
        return None

    logger.debug('cannot resolve return AST node into a known type')
    return None



from __future__ import annotations

from dataclasses import dataclass, replace
from typing import TYPE_CHECKING


try:
    from functools import cached_property
except ImportError:
    cached_property = property  # type: ignore


if TYPE_CHECKING:
    from ._ass import Ass


UNION = 'Union'


@dataclass(frozen=True)
class Type:
    """The type of a Python expression.

    It is currently limited to what can be represented by type annotations.
    """
    _name: str
    _args: list[Type]   # arguments of the type for generic types
    _ass: set[Ass]      # assumptions that  were made to infer the type
    _module: str        # the module where the type is defined, empty for built-ins

    @classmethod
    def new(
        cls,
        name: str, *,
        args: list[Type] | None = None,
        ass: set[Ass] | None = None,
        module: str = "",
    ):
        """Construct a new Type.
        """
        return cls(
            _name=name,
            _args=args or [],
            _ass=ass or set(),
            _module=module,
        )

    @property
    def name(self) -> str:
        """The name of the type.

        For example, `Iterable` or `list`.
        """
        return self._name

    @property
    def args(self) -> tuple[Type, ...]:
        """Arguments of a generic type if any.

        For example, `(str, int)` if the type is `dict[str, int]`.
        """
        return tuple(self._args)

    @property
    def module(self) -> str:
        """The module where the type is defined.

        For example, `typing` if the type is `Iterable`.
        Empty string for built-ins.
        """
        return self._module

    @cached_property
    def imports(self) -> frozenset[str]:
        """Import statements required to define the type.
        """
        result = set()
        if self._module:
            result.add(f'from {self.module} import {self._name}')
        for arg in self._args:
            result.update(arg.imports)
        return frozenset(result)

    @cached_property
    def assumptions(self) -> frozenset[Ass]:
        """Assumptions that were made when inferring the type.
        """
        result = set()
        result.update(self._ass)
        for arg in self._args:
            result.update(arg.assumptions)
        return frozenset(result)

    @property
    def unknown(self) -> bool:
        """
        infer-types can create Type with empty name.
        It is used to denote an unknown type.
        """
        return not self._name

    @property
    def signature(self) -> str:
        """Alias for `Type.annotation`.
        """
        return self.annotation

    @cached_property
    def annotation(self) -> str:
        """Represent the type as a string suitable for type annotations.

        The string is a valid Python 3.10 expression.
        For example, `str | dict[str, Any]`.
        """
        if self.unknown:
            return 'Any'
        if self.name == UNION:
            return ' | '.join(arg.annotation for arg in self._args)
        if self._args:
            args = ', '.join(arg.annotation for arg in self._args)
            return f'{self._name}[{args}]'
        return self._name

    def merge(self, other: Type) -> Type:
        """Get a union of the two given types.

        If any of the types is unknown, the other is returned.
        When possible, the type is simplified. For instance, `int | int` will be
        simplified to just `int`.
        """
        if self.unknown:
            return other
        if other.unknown:
            return self
        if self.same_as(other):
            return self
        if {self.name, other.name} == {'int', 'float'}:
            return type(self).new(
                name='float',
                ass=self._ass and other._ass,
            )
        if self.name == 'None':
            args = [other, self]
        else:
            args = [self, other]
        return type(self).new(
            name=UNION,
            args=args,
        )

    def add_ass(self, ass: Ass) -> Type:
        """Get a copy of the Type with the given Ass added in the list of assertions.
        """
        return replace(self, _ass=self._ass | {ass})

    def same_as(self, other: Type) -> bool:
        if self.name != other.name:
            return False
        if self.module != other.module:
            return False
        if self.args != other.args:
            return False
        return True
